@@ -3,9 +3,9 @@ This module provies the `~sunpy.timeseries.TimeSeriesFactory` class.
 """
 import os
 import copy
-import glob
+import pathlib
 from collections import OrderedDict
-from urllib.request import urlopen
+from urllib.request import Request
 
 import numpy as np
 import pandas as pd
@@ -17,8 +17,8 @@ from astropy.table import Table
 from astropy.time import Time
 
 import sunpy
-from sunpy.io.file_tools import UnrecognizedFileTypeError, read_file
-from sunpy.io.fits import HDPair
+from sunpy.io._fits import HDPair
+from sunpy.io.file_tools import UnrecognizedFileTypeError, detect_filetype, read_file
 from sunpy.io.header import FileHeader
 from sunpy.timeseries.timeseriesbase import GenericTimeSeries
 from sunpy.util import expand_list
@@ -29,6 +29,8 @@ from sunpy.util.datatype_factory_base import (
     NoMatchError,
     ValidationFunctionError,
 )
+from sunpy.util.functools import seconddispatch
+from sunpy.util.io import is_url, parse_path, possibly_a_path
 from sunpy.util.metadata import MetaDict
 from sunpy.util.net import download_file
 
@@ -132,13 +134,19 @@ class TimeSeriesFactory(BasicRegistrationFactory):
 
         Returns
         -------
-        parsed : `bool`
-            `True` if file has been read.
         pairs : `list` or `str`
-            List of ``(data, header)`` pairs if ``parsed`` is `True`,  ``fname`` if ``parsed`` is `False`.
-            `False` if the file is not supported or incorrect.
+            List of ``(data, header)`` pairs or ``fname`` if the file is not supported or incorrect.
         """
         if 'source' not in kwargs.keys() or not kwargs['source']:
+            try:
+                if detect_filetype(fname) == 'cdf':
+                    # Put import here to ensure there is no import dependency
+                    # on cdflib for TimeSeries
+                    from sunpy.io.cdf import read_cdf
+                    return read_cdf(os.fspath(fname), **kwargs)
+            except UnrecognizedFileTypeError:
+                pass
+
             try:
                 pairs = read_file(os.fspath(fname), **kwargs)
 
@@ -149,53 +157,34 @@ class TimeSeriesFactory(BasicRegistrationFactory):
                         data = filedata
                         meta = MetaDict(filemeta)
                         new_pairs.append(HDPair(data, meta))
-                return True, new_pairs
+                return [new_pairs]
             except UnrecognizedFileTypeError:
-                return False, fname
+                return [fname]
         else:
-            return False, fname
+            return [fname]
 
     @staticmethod
-    def _validate_meta(meta):
+    def _is_metadata(meta):
         """
-        Validate a meta argument for use as metadata.
-
-        Currently only validates by class.
+        Return `True` if ``meta`` is an object that could store metadata.
         """
-        if isinstance(meta, astropy.io.fits.header.Header):
-            return True
-        elif isinstance(meta, sunpy.io.header.FileHeader):
-            return True
-        elif isinstance(meta, dict):
-            return True
-        elif isinstance(meta, sunpy.timeseries.TimeSeriesMetaData):
-            return True
-        else:
-            return False
+        return isinstance(meta, (astropy.io.fits.header.Header,
+                                 sunpy.io.header.FileHeader,
+                                 dict,
+                                 sunpy.timeseries.TimeSeriesMetaData))
 
     @staticmethod
-    def _validate_units(units):
+    def _is_units(units):
         """
-        Validates the astropy unit-information associated with a
-        `~sunpy.timeseries.TimeSeries`.
+        Return `True` if ``units`` is an object that could store units.
 
-        Should be a dictionary of some form (but not
-        `sunpy.util.metadict.MetaDict`) with only `astropy.units` for
-        values.
+        Should be a dictionary of some form (but not `sunpy.util.metadict.MetaDict`)
+        with only `astropy.units` for values.
         """
-        result = True
-
-        # It must be a dictionary
-        if not isinstance(units, dict) or isinstance(units, MetaDict):
-            return False
-
-        for key in units:
-            if not isinstance(units[key], u.UnitBase):
-                # If this is not a unit then this can't be a valid units dict.
-                return False
-
-        # Passed all the tests
-        return result
+        # Must be a dict and all items must be a unit
+        return (isinstance(units, dict)
+                and not isinstance(units, MetaDict)
+                and all(isinstance(units[key], u.UnitBase) for key in units))
 
     @staticmethod
     def _from_table(t):
@@ -205,7 +194,7 @@ class TimeSeriesFactory(BasicRegistrationFactory):
 
         Parameters
         ----------
-        t: `~astropy.table.Table`
+        t : `~astropy.table.Table`
             The input table. The datetime column must be the first column or the (single) primary key index.
         """
         table = copy.deepcopy(t)
@@ -239,6 +228,74 @@ class TimeSeriesFactory(BasicRegistrationFactory):
         df = pd.DataFrame(data=data, index=index)
         return df, MetaDict(table.meta), units
 
+    def _parse_meta(self, meta):
+        """
+        Parse different metadata objects into a MetaDict.
+        """
+        if isinstance(meta, astropy.io.fits.header.Header):
+            meta = MetaDict(sunpy.io.header.FileHeader(meta))
+        if isinstance(meta, sunpy.timeseries.TimeSeriesMetaData):
+            new_meta = MetaDict()
+            for m in meta.metas:
+                new_meta.update(m)
+            meta = new_meta
+        return meta
+
+    def _sanitise_args(self, args):
+        """
+        Sanitise a list of args so that a single argument corresponds to either:
+
+        - (data, header, units) tuple.
+        - path-like `pathlib.Path` (e.g. a filename, directory, glob etc.).
+        - `urllib.request.Request`.
+        - `GenericTimeSeries`.
+        """
+        # Account for nested lists of items. Simply outputs a single list of
+        # items, nested lists are expanded to element level.
+        args = expand_list(args)
+
+        # Sanitise the input so that each 'type' of input corresponds to a different
+        # class, so single dispatch can be used later
+        i = 0
+        while i < len(args):
+            arg = args[i]
+            if isinstance(arg, (np.ndarray, Table, pd.DataFrame)):
+                # Extract data and metadata
+                # The next item is data
+                data = args[i]
+                meta = MetaDict()
+                units = OrderedDict()
+                if isinstance(data, Table):
+                    # We have an Astropy Table:
+                    data, new_meta, new_units = self._from_table(data)
+                    units.update(new_units)
+                    meta.update(new_meta)
+                elif isinstance(data, np.ndarray):
+                    # We have a numpy ndarray. We assume the first column is a dt index
+                    data = pd.DataFrame(data=data[:, 1:], index=Time(data[:, 0]))
+
+                # The next two could be metadata or units
+                for _ in range(2):
+                    j = i+1
+                    if j < len(args):
+                        arg = args[j]
+                        if self._is_units(arg):
+                            units.update(arg)
+                            args.pop(j)
+                        elif self._is_metadata(arg):
+                            meta.update(self._parse_meta(arg))
+                            args.pop(j)
+
+                args[i] = (data, meta, units)
+
+            elif isinstance(arg, str) and is_url(arg):
+                args[i] = Request(arg)
+            elif possibly_a_path(arg):
+                args[i] = pathlib.Path(arg)
+            i += 1
+
+        return args
+
     def _parse_args(self, *args, **kwargs):
         """
         Parses an `args` list for data-header pairs. `args` can contain any mixture of the following
@@ -264,116 +321,90 @@ class TimeSeriesFactory(BasicRegistrationFactory):
                          'directory1',
                          '*.fits')
         """
-        data_header_unit_tuples = list()
-        data_header_pairs = list()
-        already_timeseries = list()
-        filepaths = list()
+        args = self._sanitise_args(args)
+        all_ts = []
+        for arg in args:
+            try:
+                all_ts += self._parse_arg(arg, **kwargs)
+            except (NoMatchError, MultipleMatchError, ValidationFunctionError):
+                if self.silence_errors:
+                    continue
+                else:
+                    raise
+            except Exception:
+                raise
 
-        # Account for nested lists of items. Simply outputs a single list of
-        # items, nested lists are expanded to element level.
-        args = expand_list(args)
+        return all_ts
 
-        # For each of the arguments, handle each of the cases
-        i = 0
-        while i < len(args):
-            arg = args[i]
+    @seconddispatch
+    def _parse_arg(self, arg, **kwargs):
+        """
+        Parse a single arg and return a list of timeseries.
+        """
+        raise NoMatchError("File not found or invalid input")
 
-            # Data-header pair in a tuple
-            if (isinstance(arg, (np.ndarray, Table, pd.DataFrame))):
-                # and self._validate_meta(args[i+1])):
-                # Assume a Pandas Dataframe is given
-                data = arg
-                units = OrderedDict()
-                meta = MetaDict()
+    @_parse_arg.register(GenericTimeSeries)
+    def _parse_ts(self, ts, **kwargs):
+        return [ts]
 
-                # Convert the data argument into a Pandas DataFrame if needed.
-                if isinstance(data, Table):
-                    # We have an Astropy Table:
-                    data, meta, units = self._from_table(data)
-                elif isinstance(data, np.ndarray):
-                    # We have a numpy ndarray. We assume the first column is a dt index
-                    data = pd.DataFrame(data=data[:, 1:], index=Time(data[:, 0]))
+    @_parse_arg.register(Request)
+    def _parse_url(self, request, **kwargs):
+        path = download_file(request.full_url, get_and_create_download_dir())
+        return self._parse_path(pathlib.Path(path), **kwargs)
 
-                # If there are 1 or 2 more arguments:
-                for _ in range(2):
-                    if (len(args) > i+1):
-                        # If that next argument isn't data but is metaddata or units:
-                        if not isinstance(args[i+1], (np.ndarray, Table, pd.DataFrame)):
-                            if self._validate_units(args[i+1]):
-                                units.update(args[i+1])
-                                i += 1  # an extra increment to account for the units
-                            elif self._validate_meta(args[i+1]):
-                                # if we have an astropy.io FITS header then convert
-                                # to preserve multi-line comments
-                                if isinstance(args[i+1], astropy.io.fits.header.Header):
-                                    args[i+1] = MetaDict(sunpy.io.header.FileHeader(args[i+1]))
-                                if isinstance(args[i+1], sunpy.timeseries.TimeSeriesMetaData):
-                                    for j in args[i+1].metas:
-                                        meta.update(j)
-                                else:
-                                    meta.update(args[i+1])
-                                i += 1  # an extra increment to account for the meta
-
-                # Add a 3-tuple for this TimeSeries.
-                data_header_unit_tuples.append((data, meta, units))
-
-            # Filepath
-            elif (isinstance(arg, str) and
-                  os.path.isfile(os.path.expanduser(arg))):
-
-                path = os.path.expanduser(arg)
-                result = self._read_file(path, **kwargs)
-                data_header_pairs, filepaths = _apply_result(data_header_pairs, filepaths, result)
-
-            # pathlib.Path
-            elif (isinstance(arg, os.PathLike)):
-                result = self._read_file(arg.expanduser())
-                data_header_pairs, filepaths = _apply_result(data_header_pairs, filepaths, result)
-
-            # Directory
-            elif (isinstance(arg, str) and
-                  os.path.isdir(os.path.expanduser(arg))):
-
-                path = os.path.expanduser(arg)
-                files = [os.path.join(path, elem) for elem in os.listdir(path)]
-                for afile in files:
-                    # returns a boolean telling us if it were read and either a
-                    # tuple or the original filepath for reading by a source
-                    result = self._read_file(afile, **kwargs)
-                    data_header_pairs, filepaths = _apply_result(data_header_pairs, filepaths,
-                                                                 result)
-
-            # Glob
-            elif isinstance(arg, str) and '*' in arg:
-
-                files = glob.glob(os.path.expanduser(arg))
-                for afile in files:
-                    # returns a boolean telling us if it were read and either a
-                    # tuple or the original filepath for reading by a source
-                    result = self._read_file(afile, **kwargs)
-                    data_header_pairs, filepaths = _apply_result(data_header_pairs, filepaths,
-                                                                 result)
-
-            # Already a TimeSeries
-            elif isinstance(arg, GenericTimeSeries):
-                already_timeseries.append(arg)
-
-            # A URL
-            elif (isinstance(arg, str) and
-                  _is_url(arg)):
-                url = arg
-                path = download_file(url, get_and_create_download_dir())
-                result = self._read_file(path, **kwargs)
-                data_header_pairs, filepaths = _apply_result(data_header_pairs, filepaths, result)
+    @_parse_arg.register(pathlib.Path)
+    def _parse_path(self, path, **kwargs):
+        results = parse_path(path, self._read_file)
+        all_ts = []
 
+        # r can be either a TimeSeries, path, or a data, header pair
+        for r in results:
+            if isinstance(r, GenericTimeSeries):
+                all_ts += [r]
+            elif isinstance(r, pathlib.Path):
+                all_ts += [self._check_registered_widgets(filepath=r, **kwargs)]
             else:
-                raise NoMatchError("File not found or invalid input")
-            i += 1
+                pairs = r
+                # Pairs may be x long where x is the number of HDUs in the file.
+                headers = [pair.header for pair in pairs]
 
-        # TODO:
-        # In the end, if there are already TimeSeries it should be put in the
-        # same order as the input, currently they are not.
-        return data_header_unit_tuples, data_header_pairs, already_timeseries, filepaths
+                types = []
+                for header in headers:
+                    try:
+                        match = self._get_matching_widget(meta=header, **kwargs)
+                        if not match == GenericTimeSeries:
+                            types.append(match)
+                    except (MultipleMatchError, NoMatchError):
+                        continue
+
+                if not types:
+                    # If no specific classes have been found we can read the data
+                    # if we only have one data header pair:
+                    if len(pairs) == 1:
+                        all_ts += [GenericTimeSeries(pairs[0]._data, pairs[0].header)]
+                        continue
+                    else:
+                        raise NoMatchError("Input read by sunpy.io can not find a "
+                                           "matching class for reading multiple HDUs")
+                if len(set(types)) > 1:
+                    raise MultipleMatchError("Multiple HDUs return multiple matching classes.")
+
+                cls = types[0]
+
+                data_header_unit_tuple = cls._parse_hdus(pairs)
+                all_ts += self._parse_arg(data_header_unit_tuple)
+
+        return all_ts
+
+    @_parse_arg.register(tuple)
+    def _parse_tuple(self, tup, **kwargs):
+        data, header, units = tup
+        # Make a MetaDict from various input types
+        meta = header
+        if isinstance(meta, astropy.io.fits.header.Header):
+            meta = sunpy.io.header.FileHeader(meta)
+        meta = MetaDict(meta)
+        return [self._check_registered_widgets(data=data, meta=meta, units=units, **kwargs)]
 
     def __call__(self, *args, silence_errors=False, **kwargs):
         """
@@ -395,76 +426,8 @@ class TimeSeriesFactory(BasicRegistrationFactory):
         -----
         Extra keyword arguments are passed through to `sunpy.io.read_file` such as `memmap` for FITS files.
         """
-        (data_header_unit_tuples, data_header_pairs,
-         already_timeseries, filepaths) = self._parse_args(*args, **kwargs)
-
-        new_timeseries = list()
-
-        # The filepaths for unreadable files
-        for filepath in filepaths:
-            try:
-                new_ts = self._check_registered_widgets(filepath=filepath, **kwargs)
-                new_timeseries.append(new_ts)
-            except (NoMatchError, MultipleMatchError, ValidationFunctionError):
-                if not silence_errors:
-                    raise
-            except Exception:
-                raise
-
-        # data_header_pairs is a list of HDUs as read by sunpy.io
-        # For each set of HDus find the matching class and read the
-        # data_header_unit_tuples by calling the _parse_hdus method
-        # of the class.
-        for pairs in data_header_pairs:
-            # Pairs may be x long where x is the number of HDUs in the file.
-            headers = [pair.header for pair in pairs]
-
-            types = []
-            for header in headers:
-                try:
-                    match = self._get_matching_widget(meta=header, **kwargs)
-                    if not match == GenericTimeSeries:
-                        types.append(match)
-                except (MultipleMatchError, NoMatchError):
-                    continue
-
-            if not types:
-                # If no specific classes have been found we can read the data
-                # if we only have one data header pair:
-                if len(pairs) == 1:
-                    already_timeseries.append(GenericTimeSeries(pairs[0]._data,
-                                                                pairs[0].header))
-                else:
-                    raise NoMatchError("Input read by sunpy.io can not find a "
-                                       "matching class for reading multiple HDUs")
-            if len(set(types)) > 1:
-                raise MultipleMatchError("Multiple HDUs return multiple matching classes.")
-
-            cls = types[0]
-
-            data_header_unit_tuples.append(cls._parse_hdus(pairs))
-
-        # Loop over each registered type and check to see if WidgetType
-        # matches the arguments.  If it does, use that type
-        for triple in data_header_unit_tuples:
-            data, header, units = triple
-            # Make a MetaDict from various input types
-            meta = header
-            if isinstance(meta, astropy.io.fits.header.Header):
-                meta = sunpy.io.header.FileHeader(meta)
-            meta = MetaDict(meta)
-
-            try:
-                new_ts = self._check_registered_widgets(data=data, meta=meta,
-                                                        units=units, **kwargs)
-                new_timeseries.append(new_ts)
-            except (NoMatchError, MultipleMatchError, ValidationFunctionError):
-                if not silence_errors:
-                    raise
-            except Exception:
-                raise
-
-        new_timeseries += already_timeseries
+        self.silence_errors = silence_errors
+        new_timeseries = self._parse_args(*args, **kwargs)
 
         # Concatenate the timeseries into one if specified.
         concatenate = kwargs.get('concatenate', False)
@@ -476,10 +439,6 @@ class TimeSeriesFactory(BasicRegistrationFactory):
 
             new_timeseries = [full_timeseries]
 
-        # Sanitize any units OrderedDict details
-        for timeseries in new_timeseries:
-            timeseries._sanitize_units()
-
         # Only return single time series, not in a list if we only have one.
         if len(new_timeseries) == 1:
             return new_timeseries[0]
@@ -542,14 +501,6 @@ def _apply_result(data_header_pairs, filepaths, result):
     return data_header_pairs, filepaths
 
 
-def _is_url(arg):
-    try:
-        urlopen(arg)
-    except Exception:
-        return False
-    return True
-
-
 class InvalidTimeSeriesInput(ValueError):
     """
     Exception to raise when input variable is not a
